local M = {}

local scenarioFiles = {}
local championshipsUnlocked = {}
local themesUnlocked = {true, true, true, true, true}
local loaded = false
local medals = {}
local championshipData = {}
local level = {status = {state = "roaming", data = {}, position = ""}, teleport = {destination = "", championship = false, active = false, cooldown = 3, radius = 4+2, angle = 0}}
local fileHandler = require('scenario/cargames/fileHandler')
local championshipHandler = require('scenario/cargames/championshipHandler')
local particleMotion = require('scenario/cargames/particleMotion')
local helper = require("scenario/scenariohelper")

--------------------------------------------------------------------------------------------------------------------
--Misc Functions--------------------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------------------------

local function themeRequirements(theme)
	--[[
	Returns the medal requirements for each theme, or the medal requirements for a
	particular theme if specified
	]]--
	-- local themeReq = {0, 5, 10, 20, 30}
	-- Revised
	local themeReq = {0, 10, 25, 30, 35}

	if theme then
		if theme > 0 and theme <= fileHandler.getTotalNumberOfThemes() then
			return themeReq[theme]
		else
			log("E", "CarGames: hubWorld.themeRequirements()", "Cannot get the requirements of a theme that is out of bounds! [theme="..theme.."]")
		end
	else
		return themeReq
	end
end

local function getProgressOfTheme(theme)
	--[[
	Returns the completion percentage of a given theme
	Every medal counts for 3.8% each, for a total of 95% where the championship counts as the last 5%
	]]--
	local completionRate = 0

	--If the championship has been completed, 100% completion must have been achieved,
	--as all medals must have been achieved to open championship
	if championshipData.themes[theme] > 75 then
		completionRate = 100
	else
		--There are five tiers of medals over five events
		--If all are achieved, then total percentage = 95%
		--Thus, each medal is worth 95/25% = 3.8%
		local medalPercentage = 3.8
		for i = 1, 5 do
			local medalAchieved = medals[theme][i]
			local medalValue = fileHandler.getMedalRanking(medalAchieved)
			completionRate = completionRate+medalValue
		end

		completionRate = math.floor(completionRate*3.8)
	end

	return completionRate
end

local function getThemeData(themeNum)
	--[[
	Returns the theme data for a given theme
	]]--
	local completionRate = 0
	local overviewData = {}
	--Overview is treated as theme 0
	if themeNum == 0 then
		for i = 1, fileHandler.getTotalNumberOfThemes() do
			local themeCompletion = getProgressOfTheme(i)
			completionRate = completionRate+themeCompletion
		end

		completionRate = math.floor(completionRate/fileHandler.getTotalNumberOfThemes())
		table.insert(overviewData, {label = "progress", value = tostring(completionRate).."%"})
		overviewName = "Car Games: Overview"
	else
		local difficulty = 0
		for i = 1, 5 do
			local scenario = scenarioFiles[i][themeNum]
			local scenarioFile = jsonReadFile(scenario)
			local scenarioDifficulty = tonumber(scenarioFile[1]["cargames_difficulty"])
			difficulty = difficulty + scenarioDifficulty
		end

		--Get average difficulty
		difficulty = math.floor(difficulty/5)

		local completionRate = getProgressOfTheme(themeNum)

		if not themesUnlocked[themeNum] then
			table.insert(overviewData, {label = "requires", value = tostring(themeRequirements(themeNum) - fileHandler.getTotalNumberOfMedalsAchieved()).." more medals"})
		else
			table.insert(overviewData, {label = "difficulty", value = tostring(difficulty).."/10"})
			table.insert(overviewData, {label = "progress", value = tostring(completionRate).."%"})
		end

		overviewName = "Theme "..themeNum..": "..fileHandler.getThemeName(themeNum)
	end

	return overviewName, overviewData
end

local function getChampionshipData(championshipNum)
	--[[
	Returns the championship data for a given championship
	]]--
	local scenario = scenarioFiles[1][championshipNum]
	local scenarioFile = jsonReadFile(scenario)
	overviewName = scenarioFile[1]["name"]
	local score = championshipData.themes[championshipNum]

	local status
	if championshipData.data.overallScore ~= 0 then
		status = "In Progress"
	elseif score < 75 then
		status = "Incomplete"
	else
		status = "Completed"
	end

	local overviewData = {}
	table.insert(overviewData, {label = "status", value = status})
	table.insert(overviewData, {label = "best score", value = tostring(score)})

	return overviewName, overviewData
end

local function getEventData(themeNum, eventNum)
	--[[
	Returns the event data for a given event and theme number
	]]--
	local scenario = scenarioFiles[eventNum][themeNum]
	local scenarioFile = jsonReadFile(scenario)
	local score = fileHandler.getAttributeFromSave({theme = themeNum, event = eventNum, value = "score"})

	local overviewData = {}
	table.insert(overviewData, {label = "difficulty", value = scenarioFile[1]["cargames_difficulty"].."/10"})
	table.insert(overviewData, {label = "vehicle", value = scenarioFile[1]["cargames_vehicle"]})
	table.insert(overviewData, {label = "powerups", value = scenarioFile[1]["cargames_powerups"]})
	table.insert(overviewData, {label = "best score", value = tostring(score)})

	return scenarioFile[1]["name"], overviewData
end

local function getScenarioData(themeNum, eventNum, championshipNum)
	--[[
	Gets the scenario data for a given theme and event number or championship number
	]]--
	if themeNum and not eventNum then
		return getThemeData(themeNum)
	elseif championshipNum then
		return getChampionshipData(championshipNum)
	elseif themeNum and eventNum then
		return getEventData(themeNum, eventNum)
	else
		assert("Error: themeNum and/or eventNum is nil =", themeNum, " and eventNum =", eventNum)
	end
end

local function displayThemeCompletion(themeNum)
	--[[
	Displays the diamond outlines to a completed theme themeNum
	]]--
	local themeCoords = {[1] = {-174.270004, 75.9599991, 4.81}, [2] = {-21.3509998, 188.899994, 4.8}, [3] = {152.917999, 112.940002, 4.8}, [4] = {174.270004, -75.9599991, 4.96}, [5] = {21.3500004, -188.899994, 4.98}}
	TorqueScript.eval('accent'..themeNum..'.position = "'..themeCoords[themeNum][1]..' '..themeCoords[themeNum][2]..' '..themeCoords[themeNum][3]..'";')
	be:reloadStaticCollision();
end

local function lockThemes(themeNum)
	--[[
	Locks a given theme and all upcoming themes
	]]--
	--TODO Hardcoded currently - use rotation functions?
	local barrierCoords = {[2] = {9.7, 135, 9.75}, [3] = {121.706, 58.9473, 10}, [4] = {111.976, -76.099, 10}, [5] = {-9.6726, -134.806, 10}}

	for i = themeNum, fileHandler.getTotalNumberOfThemes() do
		themesUnlocked[i] = false
		TorqueScript.eval('barrier'..(i-1)..'.position = "'..barrierCoords[i][1]..' '..barrierCoords[i][2]..' '..barrierCoords[i][3]..'";')
	end

	be:reloadStaticCollision();
end

local function checkThemeLocking()
	--[[
	Checks which themes, if any, should be locked
	]]--
	local medalsAchieved = fileHandler.getTotalNumberOfMedalsAchieved()
	local numThemes = fileHandler.getTotalNumberOfThemes()

	local requirements = themeRequirements(nil)

	local lockedTheme
	for i = 2, numThemes do
		--Theme needs more medals to be unlocked
		if medalsAchieved < requirements[i] then
			lockedTheme = i
			break
		end
	end

	--If the locked theme exists, then lock it and all subsequent ones
	if lockedTheme and lockedTheme <= numThemes then
		lockThemes(lockedTheme)
	end
end

local function getNumberOfActualMedals()
	--[[
	Get a list of actual medals (ie. if diamond achieved then include the preceeding medals aswell) from savefile
	]]--
	--If player has achieved a diamond medal, then they have also achieved every medal below it as well
	local medals = fileHandler.getMedalsAchieved()

	medals.bronze   = medals.bronze  +medals.silver  +medals.gold    +medals.platinum+medals.diamond
	medals.silver   = medals.silver  +medals.gold    +medals.platinum+medals.diamond
	medals.gold     = medals.gold    +medals.platinum+medals.diamond
	medals.platinum = medals.platinum+medals.diamond
	medals.total    = medals.bronze  +medals.silver  +medals.gold    +medals.platinum+medals.diamond

	return medals
end

function getProgressStats()
	--[[
	Gets overview of stats, activated when player is in center of platform
	]]--
	local fullMedals = getNumberOfActualMedals()

	local chamionshipsCompleted = 0
	for i = 1, 5 do
		if championshipData.themes[i] > 75 then chamionshipsCompleted = chamionshipsCompleted+1 end
	end

	local progressData = {}
	table.insert(progressData, {label = "total medals", value = tostring(fullMedals.total).."/"..fileHandler.getTotalNumberOfScenarios()*5})
	table.insert(progressData, {label = ">diamond", value = tostring(fullMedals.diamond).."/"..fileHandler.getTotalNumberOfScenarios()})
	table.insert(progressData, {label = ">platinum", value = tostring(fullMedals.platinum).."/"..fileHandler.getTotalNumberOfScenarios()})
	table.insert(progressData, {label = ">gold", value = tostring(fullMedals.gold).."/"..fileHandler.getTotalNumberOfScenarios()})
	table.insert(progressData, {label = ">silver", value = tostring(fullMedals.silver).."/"..fileHandler.getTotalNumberOfScenarios()})
	table.insert(progressData, {label = ">bronze", value = tostring(fullMedals.bronze).."/"..fileHandler.getTotalNumberOfScenarios()})
	table.insert(progressData, {label = "championships", value = chamionshipsCompleted.."/"..fileHandler.getTotalNumberOfThemes()})

	guihooks.trigger('MissionInfoUpdate', {
		title = "Progress Info",
		data = progressData,
		buttons = {{text = "Back", cmd = "displayOverviewInfo(0, nil, nil)"}}
	})
end

function displayOverviewInfo(themeNumber, eventNumber, championshipNumber)
	--[[
	Displays overview information for a theme, event or championship
	]]--
	local name, scenarioData = getScenarioData(themeNumber, eventNumber, championshipNumber)

	if championshipNumber then
		guihooks.trigger('MissionInfoUpdate', {
			title = "Championship "..championshipNumber..": "..fileHandler.getThemeName(championshipNumber),
			data = scenarioData,
		})
	elseif themeNumber == 0 then
		guihooks.trigger('MissionInfoUpdate', {
			title = name,
			data = scenarioData,
			buttons = {{text = "Progress Info", cmd = "getProgressStats()"}}
		})
	else
		guihooks.trigger('MissionInfoUpdate', {
			title = name,
			data = scenarioData,
		})
	end
end

--------------------------------------------------------------------------------------------------------------------
--Transformation Functions--------------------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------------------------

local function getTranslation(translation, direction, point)
	--[[
	Returns a translation of a point, based on the direction
	]]--
	local newPoint = {z = point.z}
	if     direction == "backwards" then
		newPoint.x = point.x-translation.x
		newPoint.y = point.y-translation.y
	elseif direction == "forwards" then
		newPoint.x = point.x+translation.x
		newPoint.y = point.y+translation.y
	end

	return newPoint
end

local function getRotation(point, angle)
	--[[
	Rotates a point around an angle with center point
	]]--
	local xn = point.x*math.cos(math.rad(angle))-point.y*math.sin(math.rad(angle))
	local yn = point.y*math.cos(math.rad(angle))+point.x*math.sin(math.rad(angle))
	return {x = xn, y = yn, z = point.z}
end

local function getTranslatedPoint(center, pointToTranslate, angle)
	--[[
	Rotates a point in space by translation:
	1) Translate to (0, 0, 0)
	2) Rotate points
	3) Translate back from (0, 0, 0)
	]]
	local translatedPoint = getTranslation(center, "backwards", pointToTranslate)
	local rotatedPoint = getRotation(translatedPoint, angle)
	local newTranslatedPoint = getTranslation(center, "forwards", rotatedPoint)
	return newTranslatedPoint
end

local function getPositionInWorld(theme, event, championship)
	--[[
	Gets the position of a theme, event or championship
	]]

	-- if not theme and not event and not championship then return end
	--
	-- --For this algorithm to work, at least one theme and one medal from said theme must be provided (positions)
	-- local theme1 = {center = {x = -89.3449, y = 157.586, z = 7.75}, medals = {e1 = {x = -155.383, y = 130.627}}}
	--
	-- local center = theme1.center
	-- if theme ~= 1 then
	-- 	local themeAngle = 60*(1-(theme-1))
	-- 	center = getRotation(center, themeAngle)
	-- end
	--
	-- if championship and championship > 0 then
	-- 	--Championship center is a little above the theme center
	-- 	center.z = center.z+4
	-- 	return center
	-- elseif theme and event then
	-- 	local eventAngle = 45*(1-(event-1))
	-- 	center = getRotation(center, eventAngle)
	-- 	return center
	-- elseif theme and not event then
	-- 	return center
	-- end

	if theme and theme ~= 0 and event and event ~= 0 then
		return scenetree.findObject("themeevent"..theme..event):getPosition()
	elseif theme == 0 and event == 0 and championship ~= 0 then
		return scenetree.findObject("championshipEntry"..championship.."1"):getPosition()
	else
		return nil
	end
end

--------------------------------------------------------------------------------------------------------------------
--Teleport Functions--------------------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------------------------

local function placePlayer(data)
	--[[
	Places the player at the last scenario entered
	]]--
	--Default to middle of platform ie. don't move player from that in prefab
	if data == nil or data.theme == nil or data.event == nil then
		return
	end

	--Updates status of position
	if data.theme > 0 and data.event > 0 and data.championship == 0 then
		level.status.position = tostring(data.theme)..data.event
	elseif data.theme == 0 and data.event == 0 and data.championship > 0 then
		level.status.position = 'championship'..data.championship
	end

	local positionToPlacePlayer = getPositionInWorld(data.theme, data.event, data.championship)

	local z = 0
	if data.theme == 1 or data.theme == 3 or data.championship == 1 or data.championship == 3 then
		z = 9.9561
	elseif data.theme == 2 or data.championship == 2 then
		z = 9.83196
	elseif data.theme == 4 or data.championship == 4 then
		z = 10.0856
	elseif data.theme == 5 or data.championship == 5 then
		z = 10.059
	end

	--Place player on the higher platform
	if data.championship and data.championship ~= 0 then
		z = z+1.75
	end

	if positionToPlacePlayer then
		TorqueScript.eval('scenario_player0.position = "'..positionToPlacePlayer.x..' '..positionToPlacePlayer.y..' '..z..'";')
	end
end

local function teleportToEvent(theme, event)
	--[[
	Loads up the event selected by player
	]]--
	--Prevents the scenario from continuing to call this function while transistion period occurs
	level.teleport.active = false
	local filePath
	if level.teleport.championship then
		championshipHandler.saveChampionshipData(championshipData, true)
		if championshipData.data.event ~= 0 and championshipData.data.theme ~= 0 then
			-- +1 since savedata records last event completed (should never have to worry about it being > 5 since championship terminates at 5)
			filePath = scenarioFiles[championshipData.data.event+1][championshipData.data.theme]
		else
			filePath = scenarioFiles[event][theme]
		end
	else
		championshipHandler.reset()
		filePath = scenarioFiles[event][theme]
	end
	filePath = filePath:sub(2)
	guihooks.trigger("MissionInfoUpdate", nil)

	--Tells the scenario loading up that the previous level was the hub-world
	--Specifically useful for championship events
	fileHandler.setLastScenarioInfo({theme = 0, event = 0, championship = 0})

	scenario_scenariosLoader.start(scenario_scenariosLoader.loadScenario(filePath, nil, filePath))
end

local function beginTeleport(dt)
	--[[
	Starts the teleporting procedure by displaying particles and once the countdown
	has finished, teleports the player
	]]--
	level.teleport.angle = level.teleport.angle+dt*(5-level.teleport.cooldown)*6
	local x, y = scenetree.findObject("scenario_player0"):getPosition().x, scenetree.findObject("scenario_player0"):getPosition().y
	local xOff, yOff = particleMotion.circular(level.teleport.radius, level.teleport.angle)

	TorqueScript.eval('teleport.position = "'..(x+xOff)..' '..(y+yOff)..' 9.61568";')

	level.teleport.cooldown = level.teleport.cooldown-dt
	if level.teleport.cooldown <= 0 then
		local themeNumber, eventNumber = tonumber(level.teleport.destination:sub(1, 1)), tonumber(level.teleport.destination:sub(2, 2))
		teleportToEvent(themeNumber, eventNumber)
	end
end

local function cancelTeleport()
	--[[
	Prevents the teleport from going ahead
	(called when the player leaves the event teleport area)
	]]--
	if level.teleport.active then ui_message("Teleport Cancelled", 3, "A") end
	level.status.position = ""
	level.teleport.destination = ""
	level.teleport.cooldown = 3
	level.teleport.active = false
	level.teleport.angle = 0
	level.teleport.radius = 6
	level.teleport.championship = false
	TorqueScript.eval('teleport.position = "10000 0 0";')
end

--------------------------------------------------------------------------------------------------------------------
--Medal Functions--------------------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------------------------

local function createLight(colour, id)
	--[[
	Creates a light into the world
	]]--
	TorqueScript.eval([[
	new PointLight(]]..("light"..id)..[[) {
		radius = "25";
		isEnabled = "1";
		color = "]]..colour.red..[[ ]]..colour.green..[[ ]]..colour.blue..[[ 2";
		brightness = "1.5";
		castShadows = "1";
	};
	]])
end

local function createMedal(type, id)
	--[[
	Creates a medal into the world
	]]--
	TorqueScript.eval([[
	new TSStatic(medal]]..id..[[) {
		shapeName = "levels/smallgridspace/art/shapes/campaign/medals/]]..type..[[Medal.dae";
		collisionType = "Visible Mesh Final";
		decalType = "Visible Mesh Final";
	};
	]])
end

local function createCube(type, id)
	--[[
	Creates a medal cube into the world
	]]--
	TorqueScript.eval([[
	new TSStatic(cube]]..id..[[) {
		shapeName = "levels/smallgridspace/art/shapes/campaign/medals/]]..type..[[Cube.dae";
		collisionType = "Visible Mesh Final";
		decalType = "Visible Mesh Final";
	};
	]])
end

local function getColourOfMedal(medal)
	--[[
	Returns the RGB colour of a medal
	]]--
	if     medal == "bronze"   then return {red = 0.36, green = 0.28, blue = 0.07}
	elseif medal == "siler"    then return {red = 0.15, green = 0.15, blue = 0.15}
	elseif medal == "gold"     then return {red = 1   , green = 0.89, blue = 0   }
	elseif medal == "platinum" then return {red = 0.66, green = 0.73, blue = 0.74}
	elseif medal == "diamond"  then return {red = 0   , green = 0.82, blue = 1   }
	end
end

local function placeChampionships()
	--[[
	Places the championship at their respective position if they have been unlocked
	]]--
	local championships = {}
	for i = 1, fileHandler.getTotalNumberOfThemes() do
		--Assumed championship is open until proven not
		championships[i] = true
		for j = 1, 5 do
			if medals[i][j] ~= "diamond" then
				championships[i] = false
				break
			end
		end
	end

	--Check for if championship is complete and thus theme is completed fully
	for i = 1, 5 do
		if championships[i] then
			local score = championshipData.themes[i]
			if score >= 75 then
				displayThemeCompletion(i)
			end
		end
	end

	local platforms = {theme1 = {x = -88.6326, y = 136.98}}
	local angle = 0
	for i = 2, 5 do
		angle = angle-60
		local newRotatedPoint = getRotation(platforms.theme1, angle)
		platforms["theme"..i] = newRotatedPoint
	end

	for j = 1, 5 do
		if championships[j] then
			TorqueScript.eval('championship'..j..'.position = "'..platforms["theme"..j].x..' '..platforms["theme"..j].y..' 9.5";')
			TorqueScript.eval('championship'..j..'.rotation = "0 0 -1 '..(30-(j-1)*60)..'";')
		end
	end

	be:reloadStaticCollision();

	championshipsUnlocked = championships
end

local function placeMedals()
	--[[
	Places the medals for each event
	]]--

	local themes = {}
	--For this algorithm to work, at least one theme and one medal from said theme must be provided (positions)
	local theme1 = {center = {x = -89.3449, y = 157.586}, medals = {e1 = {x = -155.383, y = 130.627}}, lights = {e1 = {x = -152.934, y = 135.104}}, cubes = {c1 = {x = -111.76, y = 78.47, z = 12.75}, c2 = {x = -9.73, y = 136.19, z = 12.75}}}

	local angle = 0

	--All z values are the same
	local z = theme1.cubes.c1.z
	--Theme 1 is set some only worry about 2-5
	for i = 2, 5 do
		angle = angle-44.95

		theme1.medals["e"..i] = getTranslatedPoint(theme1.center, theme1.medals.e1, angle)
		theme1.lights["e"..i] = getTranslatedPoint(theme1.center, theme1.lights.e1, angle)

		z = z+6
		theme1.cubes["c"..((i-1)*2)+1] = {x = theme1.cubes.c1.x, y = theme1.cubes.c1.y, z = z}
		theme1.cubes["c"..(((i-1)*2)+2)] = {x = theme1.cubes.c2.x, y = theme1.cubes.c2.y, z = z}
	end

	themes.theme1 = theme1

	--Loop through themes and rotate 60 degrees as each theme is rotated by the center point at the same angle
	local themeAngle = 0
	for j = 2, 5 do
		--Each theme is roated back 60 degrees
		themeAngle = themeAngle-60
		--Setup table
		themes["theme"..j] = {medals = {}, lights = {}, cubes = {}}
		--Update center
		themes["theme"..j]["center"] = getRotation(themes.theme1.center, themeAngle)

		--Medals and lights
		for k = 1, 5 do
			local newRotatedPointMedal = getRotation(themes["theme1"]["medals"]["e"..k], themeAngle)
			local newRotatedPointLight = getRotation(themes["theme1"]["lights"]["e"..k], themeAngle)
			themes["theme"..j]["medals"]["e"..k] = newRotatedPointMedal
			themes["theme"..j]["lights"]["e"..k] = newRotatedPointLight
		end

		--Cubes
		for l = 1, 10 do
			local newRotatedPoint = getRotation(themes["theme1"]["cubes"]["c"..l], themeAngle)
			themes["theme"..j]["cubes"]["c"..l] = newRotatedPoint
		end
	end

	--Create and place
	for t = 1, 5 do
		medals[t] = {}
		local lowestMedal = "diamond"
		for e = 1, 5 do
			local name = ""..t..e

			--If it exists in the level then delete it and replace it
			if scenetree.findObject("medal"..name) ~= nil then
				TorqueScript.eval("medal"..name..'.delete();')
			end

			local data = {theme = t, event = e, value = "medal"}
			local medalAchieved = fileHandler.getAttributeFromSave(data)

			--Used for championships and is more efficient to not call this twice
			if fileHandler.getMedalRanking(medalAchieved) < fileHandler.getMedalRanking(lowestMedal) then lowestMedal = medalAchieved end

			medals[t][e] = medalAchieved
			--Only place if a medal is achieved
			if medalAchieved ~= "none" then
				createMedal(medalAchieved, name)

				local colour = getColourOfMedal(medalAchieved)
				if colour and not scenetree.findObject("light"..name) then
					createLight(colour, name)
				end

				TorqueScript.eval("medal"..name..'.position = "'..themes["theme"..t]["medals"]["e"..e].x..' '..themes["theme"..t]["medals"]["e"..e].y..' 11.027";')
				--Rotation:
				--Each medal is 45 degrees rotated from each other in the same theme (starting at 15 degrees)
				--Each theme is rotated via 60 degrees so between two medals of the same event will be a 60 + 45 degree change
				local themeRotation, medalRotation = (t-1)*60, 15 + (e-1)*45
				local rotation = themeRotation + medalRotation
				TorqueScript.eval("medal"..name..'.rotation = "0 0 1 '..rotation..'";')

				TorqueScript.eval("light"..name..'.position = "'..themes["theme"..t]["lights"]["e"..e].x..' '..themes["theme"..t]["lights"]["e"..e].y..' 17";')
			end
		end

		cubeValue = {material = lowestMedal, value = fileHandler.getMedalRanking(lowestMedal)}

		if cubeValue.material ~= "none" then
			local numCubesToPlace = cubeValue.value*2

			for c = 1, numCubesToPlace do
				local name = ""..t..c

				if scenetree.findObject("cube"..name) ~= nil then TorqueScript.eval("cube"..name..'.delete();') end

				createCube(cubeValue.material, name)

				TorqueScript.eval("cube"..name..'.position = "'..themes["theme"..t]["cubes"]["c"..c].x..' '..themes["theme"..t]["cubes"]["c"..c].y..' '..themes["theme"..t]["cubes"]["c"..c].z..'";')
				local angle = 105+(t-1)*60
				if c % 2 == 0 then angle = angle+90 end
				TorqueScript.eval("cube"..name..'.rotation = "0 0 1 '..angle..'";')
			end
		end
	end
end

-------------------------------------------------------------------------------------------------------------------

local function onBeamNGTrigger(data)
	if data.triggerName:find("themeevent") or data.triggerName:find("championshipEntry") then
		if data.event == "enter" then
			local destination
			if     data.triggerName:find("themeevent")        then
				destination = data.triggerName:gsub("themeevent", "")
				level.teleport.championship = false
			elseif data.triggerName:find("championshipEntry") then
				--Gets theme
				destination = data.triggerName:gsub("championshipEntry", "")
				local championshipEvent = championshipData.data.event
				if championshipEvent ~= 0 then
					destination = destination:sub(1, -2)
					destination = destination..championshipEvent
				end
				level.teleport.championship = true
			end

			if level.status.position ~= destination and not level.status.position:find("championship") then
				level.teleport.destination = destination
				level.teleport.active = true
				ui_message("Teleporting...", 5, "A")
			end
		elseif data.event == "exit" then
			cancelTeleport()
		end
	elseif data.triggerName == "fallenOff" and data.event == "enter" then
		TorqueScript.eval('scenario_player0.position = "0.459702015 -2.59148145 9.95547771";')
		TorqueScript.eval('scenario_player0.rotation = "0 0 0 1";')
	elseif data.triggerName:find("theme") then
		if data.event == "enter" then
			if loaded then
				local theme = data.triggerName:gsub("theme", "")
				theme = tonumber(theme)
				displayOverviewInfo(theme, nil)
				level.status.state = "displayingInfo"
				level.status.data = "theme"..theme
			end
		elseif data.event == "exit" then
			--Fixes an issue where the main overview UI would be removed AFTER setting up the next one
			--causing the new one to never display
			if data.triggerName == level.status.data then
				guihooks.trigger("MissionInfoUpdate", nil)
				level.status.state = ""
			end
		end
	end
end

local function onScenarioChange(sc)
	if sc.state == "pre-running" and not loaded then
		guihooks.trigger('MissionInfoUpdate', nil)
		scenarioFiles = fileHandler.findAllFiles()
		--Setup and load championshipData from championshipHandler
		championshipData = championshipHandler.loadFullChampionshipData()

		--Update changes the savedata for championships so reset them to prevent issues
		if championshipData.data.retries then
			championshipHandler.backwardsCompatibility()
			--Reload data
			championshipData = championshipHandler.loadFullChampionshipData()
		end

		checkThemeLocking()

		placeMedals()
		placeChampionships()

		local data = fileHandler.getMiscAttribute("lastScenario")
		placePlayer(data)

		loaded = true
	end
end

local function onPreRender(dt)
	if loaded then
		if level.teleport.active then beginTeleport(dt) end

		local vehicleObject = scenetree.findObject("scenario_player0")
		if not vehicleObject then return end

		for t = 1, fileHandler.getTotalNumberOfThemes() do
			local championshipTrigger = scenetree.findObject("championshipEntry"..t..'1')
			local distanceChampionship = math.floor(math.sqrt(math.pow(vehicleObject:getPosition().x-championshipTrigger:getPosition().x, 2) + math.pow(vehicleObject:getPosition().y-championshipTrigger:getPosition().y, 2) + math.pow(vehicleObject:getPosition().z-championshipTrigger:getPosition().z, 2)))
			if distanceChampionship <= 15 and level.status.state ~= "displayingInfo" then
				if championshipsUnlocked[t] == true then
					displayOverviewInfo(nil, nil, t)
					level.status.state = "displayingInfo"
					level.status.data = "championship"..t
					break
				end
			elseif distanceChampionship > 15 and level.status.state == "displayingInfo" then
				if level.status.data == "championship"..t then
					guihooks.trigger("MissionInfoUpdate", nil)
					level.status.state = ""
					break
				end
			end
			for e = 1, 5 do
				local trigger = scenetree.findObject("themeevent"..t..""..e)
				local distance = math.floor(math.sqrt(math.pow(vehicleObject:getPosition().x-trigger:getPosition().x, 2) + math.pow(vehicleObject:getPosition().y-trigger:getPosition().y, 2) + math.pow(vehicleObject:getPosition().z-trigger:getPosition().z, 2)))
				if distance <= 15 and level.status.state ~= "displayingInfo" then
					displayOverviewInfo(t, e)
					level.status.state = "displayingInfo"
					level.status.data = ""..t..""..e
					break
				elseif distance > 15 and level.status.state == "displayingInfo" then
					if level.status.data == ""..t..""..e then
						guihooks.trigger("MissionInfoUpdate", nil)
						level.status.state = ""
						break
					end
				end
			end
		end
	end
end

local function onRaceStart()
	local data = fileHandler.getMiscAttribute("lastScenario")
	--Checks to see if the player is in the default location and if so, displays the
	--overview progress UI
	if data.theme == 0 and data.event == 0 and data.championship == 0 then
		displayOverviewInfo(data.theme, nil)
		level.status.state = "displayingInfo"
		level.status.data = "theme"..data.theme
	end
end

--Removes the UI if the scenario is unloaded; it does not linger in the next loaded scenario/level/etc
local function onClientEndMission(m) guihooks.trigger("MissionInfoUpdate", nil) end

M.onClientEndMission = onClientEndMission
M.onBeamNGTrigger = onBeamNGTrigger
M.onScenarioChange = onScenarioChange
M.onRaceStart = onRaceStart
M.onPreRender = onPreRender
return M
