local M = {}

local fileHandler = require('scenario/cargames/fileHandler')

local function loadChampionshipData()
  --[[
  Retrieves the current championship data and returns it
  ]]--
  return fileHandler.getMiscAttribute("championship").data
end

local function loadFullChampionshipData()
  --[[
  Retrieves the full amount of championship data stored in the file and returns it
  ]]--
  return fileHandler.getMiscAttribute("championship")
end

local function saveChampionshipData(championshipData, engaged)
  --[[
  --Overwrites the old data with the new, and changes state if necessary
  ]]--
  fileHandler.setChampionshipInfo(championshipData, engaged)
end

local function isInChapionshipMode()
  --[[
  Returns true if the championship is being run, else returns false
  ]]--
  return fileHandler.getMiscAttribute("championship").engaged
end

local function getMedalPointConversion(medal)
  --[[
  Returns the amount of points each medal is worth
  ]]--
  if     medal == 1 then return 10
  elseif medal == 2 then return 12
  elseif medal == 3 then return 15
  elseif medal == 4 then return 20
  elseif medal == 5 then return 25
  else return 0
  end
end

local function reset()
  --[[
  Resets all of the variables and then overwrites the old data with the new
  ]]--
  local championshipData = {}
  championshipData.event = 0
  championshipData.theme = 0
  championshipData.eventScores = {}
  championshipData.overallScore = 0
  saveChampionshipData(championshipData, false)
end

local function backwardsCompatibility()
  --[[
  Updates an older championship savefile to the new one
  ]]--
  local data = loadFullChampionshipData()
  for theme, score in ipairs(data.themes) do
    if score >= 15 then
      --Adjust the score to the new required amount
      --unfortunately, not possible to work out exact mapping of older score to new,
      --due to the way medal points have changed
      data.themes[theme] = 75
    end
  end

  saveChampionshipData(data, false)

  reset()
end

local function skipEvent()
  --[[
  Called when the user restarts a championship event
  This function skips the event an loads the next one
  If the user got a score and then reset, this score is honored, but if they
  didn't, records 0
  ]]--
  local data = loadChampionshipData()
  local eventsRecorded = #data.eventScores

  if eventsRecorded < data.event then
    --Record 0 for resetting during an event and not finishing
    table.insert(data.eventScores, 0)
  end

  saveChampionshipData(data, true)
end

local function endChampionship(champData, theme)
  --[[
  Ends the championship and saves the score to the appropiate theme
  ]]--
  local currentData = loadFullChampionshipData()
  local currentBestForTheme = currentData.themes[theme]

  --Update current best
  if champData.overallScore > currentBestForTheme then
    currentData.themes[theme] = champData.overallScore
  end

  saveChampionshipData(currentData, true)
  reset()
end

local function endEvent(scenarioData)
  --[[
  Handles the championship event ending
  ]]--
  local pts = getMedalPointConversion(scenarioData.medal)

  local champData = loadChampionshipData()

  --Check for old savefile and fix this before continuing
  if not champData.eventScores then
    backwardsCompatibility()
    --Reload data
    champData = loadChampionshipData()
  end

  champData.overallScore = champData.overallScore + pts
  table.insert(champData.eventScores, pts)

  champData.theme = scenarioData.theme
  champData.event = scenarioData.event

  saveChampionshipData(champData, true)

  --Last event so end championship
  if scenarioData.event == 5 then
    endChampionship(champData, scenarioData.theme)
    return true, champData
  end

  return false, champData
end

M.isInChapionshipMode = isInChapionshipMode
M.saveChampionshipData = saveChampionshipData
M.loadChampionshipData = loadChampionshipData
M.loadFullChampionshipData = loadFullChampionshipData
M.reset = reset
M.backwardsCompatibility = backwardsCompatibility
M.skipEvent = skipEvent
M.getMessage = getMessage
M.endChampionship = endChampionship
M.endEvent = endEvent
return M
