local M = {}

local event               = require('scenario/cargames/eventHandler')
local file                = require('scenario/cargames/fileHandler')
local helper              = require('scenario/scenariohelper')
local uiHandler 					= require('scenario/cargames/uiHandler')
local rotations           = require('scenario/cargames/rotationHandler')
local championshipHandler = require('scenario/cargames/championshipHandler')

local scenarioInstance

local function getTheme(filePath)
	--[[
	Returns the theme of the current scenario
	]]--
	return tonumber(filePath:match("%d+"))
end

local function findPowerups()
	--[[
	Finds all powerups in a scenario
	]]--
	--List of all Powerups
	local typesOfPowerups = {"x2", "x4", "x8", "juggernaut", "time5", "time10", "nitro"}
	--By convention, the powerup (in scenario) will be called <powerup>_<id>_board
	local powerupsInScenario = {}
	for i = 1, #typesOfPowerups do
		local id = 1
		while scenetree.findObject(typesOfPowerups[i].."_"..id.."_board") do
			local obj = scenetree.findObject(typesOfPowerups[i].."_"..id.."_board")
			--x, y, z scale values are ALWAYS the same and so doesn't matter which value is used (x or y or z)
			local data = {id = id, type = typesOfPowerups[i], position = obj:getPosition(), scale = obj:getScale().x*3}
			table.insert(powerupsInScenario, data)
			id = id + 1
		end
	end

	return powerupsInScenario
end

---------------------------------------------------------------------------------------------------------------------------------------------
---Class Powerup---
---------------------------------------------------------------------------------------------------------------------------------------------

Powerup = {}
Powerup.__index = Powerup

function Powerup:reset()
	--[[
	Resets powerup in world
	]]--
	TorqueScript.eval(self.name..'_'..self.id..'_board.position = "'..self.position.x..' '..self.position.y..' '..self.position.z..'";')
end

function Powerup:new(name, id, position, scale)
	local instance = {name = name, id = id, position = position, scale = scale, triggered = false}
	TorqueScript.eval(name..'_'..id..'_board.position = "'..position.x..' '..position.y..' '..position.z..'";')
	setmetatable(instance, Powerup)
	return instance
end

function Powerup:getObjectName()
	--[[
	Returns the object's name in the game engine
	]]--
	return self.name..'_'..self.id
end

function Powerup:getType()
	--[[
	Returns the object's type
	]]--
	return self.name
end

function Powerup:getId()
	--[[
	Returns the object's id
	]]--
	return self.id
end

function Powerup:hasBeenTriggered()
	--[[
	Returns whether the powerups was activated during the event
	]]
	return self.triggered
end

function Powerup:trigger()
	--[[
	Triggers the powerup by performing specific tasks that the powerup entails
	]]
	uiHandler.updatePowerups(self.name..self.id)
	TorqueScript.eval(self.name..'_'..self.id..'_board.position = "10000 0 0";')
	self.triggered = true

	if self.name == "juggernaut" then
		local truck = scenetree.findObject('truck'..self.id)
		--Apply acceleration to it
		truck:queueLuaCommand('obj:setWind(0, 300, 0)')

		--UPDATE: The wind now causes the truck to loose control so the AI now help to control the truck
		--Set the truck on a path to keep the truck straight
		local arg = {vehicleName = 'truck'..self.id,
              	 waypoints = {'truckTarget'..self.id},
			  				 lapCount = 1,
              	 aggression = 1}

		helper.setAiPath(arg)

		if scenarioInstance ~= nil then scenarioInstance:triggerTruck(self.id) end
	elseif self.name:find("time") then
		--Find time based on name of powerup
		local timeToAdd = self.name:gsub("[^0-9]", "")
		--Call the function in the scenarioInstance's class
		if scenarioInstance ~= nil then
			scenarioInstance:addTime(timeToAdd)
		end
	elseif self.name:find("nitro") then
		if scenarioInstance ~= nil then
			scenarioInstance:addNitro(1)
		end
	end
end

function Powerup:checkDistance(vehicleObject)
	--[[
	Checks to see if the player is within triggering distance of the powerup, and
	triggers it accordingly
	]]--
	if not vehicleObject then return end
	--If the powerup is not already triggered, check to see if it needs triggering
	if not self.triggered then
		local distance = math.floor(math.sqrt(math.pow(vehicleObject:getPosition().x-self.position.x, 2) + math.pow(vehicleObject:getPosition().y-self.position.y, 2) + math.pow(vehicleObject:getPosition().z-self.position.z, 2)))
		--Distance the object needs to be in range of depends on the scale of the powerup itself
		if distance <= self.scale then
			self:trigger()
		end
	end
end

---------------------------------------------------------------------------------------------------------------------------------------------
---Class Event---
---------------------------------------------------------------------------------------------------------------------------------------------

ClassEvent = {}
ClassEvent.__index = ClassEvent

function ClassEvent.new(attributes)
	local instance = {finished = false}

  for k, v in pairs(attributes) do
    --Add attribute to instance
    instance[k] = v
  end

	instance.vehicles = instance.vehicles or {"scenario_player0"}

  --For these events to work, eventName must be setup
  if not instance.eventName then
    log("E", "Car Games", "Event was setup without an eventName! Loading has been terminated")
    return
  end

	local foundPowerups = findPowerups()
	--Important: This is provisionally set to nil so that events can skip code if no powerups are in it
	local powerups = nil

	for k, v in pairs(foundPowerups) do
		if not powerups then
			powerups = {instances = {}, types = {nitro = false, time = false, multiplier = false, juggernaut = false}}
		end

		--Appends which types of powerups exist in the scenario
		if not powerups.types[v.type] then
			--Determine if the type is a multiplier
			if v.type:find("x") and #v.type == 2 then
				powerups.types.multiplier = true
			else
				powerups.types[v.type] = true
				if v.type == "nitro" then
					--Sets up nitro
					instance.nitro = 0
				end
			end
		end

		table.insert(powerups.instances, Powerup:new(v.type, v.id, v.position, v.scale))
	end

	instance.powerups = powerups

	if powerups then
		--Setup the UI by giving it the list of powerups in the scenario
		uiHandler.updatePowerups(nil, powerups.instances)
	end

  --Setup the UI by updating it (prevents UI displaying null)
  uiHandler.updateMedalUI(instance.eventName, instance.theme, nil, nil, nil)

	--Handle any rotations in the event
	rotations.setupCelestialRotations()

	setmetatable(instance, ClassEvent)
	return instance
end

function ClassEvent:callMethod(func, var, ...)
	--[[
	Returns the result of a function call in this superclass or subsclass
	]]--
	return scenarioInstance[func](self, var, ...)
end

function ClassEvent:isFinished()
	--[[
	Returns whether the event has finished yet
	]]--
	return self.finished
end

function ClassEvent:addNitro(amount)
	--[[
	Adds a set of nitro to those already got (can be negative to take boost away)
	]]--
	self.nitro = self.nitro + amount
end

function ClassEvent:addTime(amount)
	--[[
	Adds an amount of time (can be negative to take away time)
	]]--
	self.scenarioTimeLimit = self.scenarioTimeLimit + amount
end

function ClassEvent:getNumberOfPowerupsInScenario(powerupType)
	--[[
	Returns the number of speicified powerups in the scenario
	]]--
	local count = 0
	for k, v in pairs(self.powerups.instances) do
		if v.name == powerupType then
			count = count + 1
		end
	end

	return count
end

function ClassEvent:onPreRender(dt)
  --Don't check for powerups if the event is finished or there are none
	if not self.finished and self.powerups ~= nil then
		for k1, powerup in pairs(self.powerups.instances) do
			for k2, car in pairs(self.vehicles) do
				powerup:checkDistance(scenetree.findObject(car))
			end
		end

		--Update nitro UI if there is some in the event
		if self.powerups.types.nitro then
			uiHandler.updateNitroUI(self:getNumberOfPowerupsInScenario("nitro"), self.nitro, dt*10, not self.active)
		end
	end

	--Update rotations
	rotations.updateRotations(dt*bullettime.get())
end

function ClassEvent:onScenarioRestarted()
	--Check if the user is in a championship
	local champData = championshipHandler.loadFullChampionshipData()
	if champData.engaged then
		--If so, skip the event, since you cannot retry the event
		championshipHandler.skipEvent()

		local next = event.getNextEvent(event.getEventNumber(self.eventName), self.theme)
		local filePath = ""

		if next then
			--Load the next scenario
			filePath = next
		else
			championshipHandler.endChampionship(champData.data, self.theme)
			--Load hub-world
			filePath = "levels/smallgridspace/scenarios/cargames/roam/cargameshub.json"
		end

		self.load = filePath
	end

	--Reset slow-motion
	helper.queueLuaCommandByName('scenario_player0', 'bullettime.set(1)')
	--Reset all powerups if there are any
	if self.powerups ~= nil then
		for k, v in pairs(self.powerups.instances) do
			v:reset()
		end
	end
end

function ClassEvent:onBeamNGTrigger(data)
	--End scenario if fallen off of course
	if data.triggerName == 'fallen' then
		self:fail("You fell off the course")
	end
end

function ClassEvent:onScenarioChange(sc)
	if sc.state == "pre-running" and self.load then
		scenario_scenariosLoader.start(scenario_scenariosLoader.loadScenario(self.load, nil, self.load))
	end
end

function ClassEvent:fail(reason)
	--[[
	Terminates the scenario and displays the end message as a failure
	The purpose of this is for single event lua files to call this to fail the scenario,
	rather than have to implement it itself (which is redundant and spatially inefficient)
	]]--
	self.finished = true
	event.endScenario({theme = self.theme, event = self.eventName, name = self.scenarioName, failed = true, msg = reason})
end

--These functions should be overwritten if they want to do anything useful
--These are here so if any are called accidently/in error, then BeamNG will not error
function ClassEvent:forceScore() end
function ClassEvent:onRaceTick() end
function ClassEvent:onRaceStart() end

---------------------------------------------------------------------------------------------------------------------------------------------
---Class Darts - Inherits/Extends off of ClassEvent---
---------------------------------------------------------------------------------------------------------------------------------------------

Darts = {};
Darts.__index = Darts

function Darts.new(eventType, filePath, scenarioName)
	local instance = ClassEvent.new({
			eventName = eventType,
			scenarioName = scenarioName,
			theme = getTheme(filePath),
			scenarioTimeOut = 0,
			scenarioTime = 0,
			scoreLand = 0,
			boardPosition = scenetree.findObject("board"):getPosition()
		}
	)
	setmetatable(instance, Darts)
	return instance
end

--Inheritance
setmetatable(Darts, {__index = ClassEvent})

--Override function
function Darts:onRaceTick(raceTickTime)
	if not self.finished then
		self.scenarioTimeOut = self.scenarioTimeOut+raceTickTime
		local playerObject = scenetree.findObject("scenario_player0")

		--If the player has not scored yet, keep checking for their landing
		if self.scoreLand == 0 then
			if playerObject:getPosition().z <= self.boardPosition.z+5 then
				self.scoreLand = 100-math.sqrt(math.pow(playerObject:getPosition().x-self.boardPosition.x, 2) + math.pow(playerObject:getPosition().y-self.boardPosition.y, 2) + math.pow(playerObject:getPosition().z-self.boardPosition.z, 2))
			end
		else
			local playerData = map.objects[map.objectNames["scenario_player0"]]
			self.scenarioTime = self.scenarioTime+raceTickTime
			--If the player is still or if the car keeps moving but exceeds the time limit, end scenario
			if self.scenarioTime >= 8 or playerData.vel:length() <= 0.025 then
				event.finishScenario(self.eventName, self.theme, self.scenarioName, self.powerups, {scoreLand = self.scoreLand, scoreRest = 100-math.sqrt(math.pow(playerObject:getPosition().x-self.boardPosition.x, 2) + math.pow(playerObject:getPosition().y-self.boardPosition.y, 2) + math.pow(playerObject:getPosition().z-self.boardPosition.z, 2))})
				self.finished = true
			end
		end
	end
end

---------------------------------------------------------------------------------------------------------------------------------------------
---Class Curling - Inherits/Extends off of ClassEvent---
---------------------------------------------------------------------------------------------------------------------------------------------

Curling = {};
Curling.__index = Curling

function Curling.new(eventType, filePath, scenarioName)
	--Gets the vehicles in the scenario
	local vehicles = {}

	local i = 0
	while scenetree.findObject("scenario_player"..i) do
		table.insert(vehicles, "scenario_player"..i)
		i = i+1
	end

	--Reset and use same variable
	i = 0
	while scenetree.findObject("opponent"..i) do
		table.insert(vehicles, "opponent"..i)
		i = i+1
	end

	--Reset and use same variable (trucks start at 1)
	i = 1
	while scenetree.findObject("truck"..i) do
		table.insert(vehicles, "truck"..i)
		i = i+1
	end

	--These are mandatory attributes
	local attributes = {
		scenarioName = scenarioName,
		eventName = eventType,
		theme = getTheme(filePath),
		forcedScores = {},
		pendTruck = 0,
		scenarioTimeOut = 0,
		startTimer = false,
		hasBegun = false,
		vehicles = vehicles
	}

	local instance = ClassEvent.new(attributes)

  --For these types of events to work, vehicles must be setup
  if not instance.vehicles then
    log("E", "Car Games", "Event was setup without any vehicle data! Loading has been terminated")
    return
  end

	setmetatable(instance, Curling)
	return instance
end

--Inheritance
setmetatable(Curling, {__index = ClassEvent})

function Curling:truckActivated()
	--[[
	Determines whether a truck is activated and in effect (ie. hasn't yet entered
	--the scoreboard) or not
	--DEPRECATED
	]]--
	if self.trucks == nil then
		return false
	else
		--If any truck has been activated, return true
		for k, v in (self.trucks)  do
			if v.triggered then
				return true
			end
		end

		return false
	end
end

function Curling:onRaceTick(raceTickTime)
end

--Override function
function Curling:onPreRender(dt)
	--Since this function has been overwritten, this is needed to access classEvent's
	--onPreRender function: ie. emulates super()
	ClassEvent.onPreRender(self, dt)

	if not self.finished then
		uiHandler.updateMedalUI(self.eventName, self.theme, self.powerups, true, nil, self.vehicles, self.forcedScores)

		if self.pendTruck == 0 then
			uiHandler.updateTimerUI(self.scenarioTimeOut, 5, self.eventName)
		else
			--If a truck is pending, reset and freeze the timer
			uiHandler.updateTimerUI(0, 5, self.eventName)
		end

		--Timeout is set to 5 seconds, trucks will restart the count back to 0
		--If the event has finished, prevent this function from continuously running
		if self.scenarioTimeOut >= 5 and not self.finished then
			helper.queueLuaCommandByName('scenario_player0', 'bullettime.set(1)')
			--If not trucks have not yet entered the scoreboard, end the scenario
			if self.pendTruck == 0 then
				event.finishScenario(self.eventName, self.theme, self.scenarioName, self.powerups, self.vehicles, self.forcedScores)
				self.finished = true
			--Indicates that a truck has been scheduled and so the timer should restart
			else
				self.scenarioTimeOut = 0
				self.startTimer = false
			end
		end

		--No point in running this if the event has not run or the scenario is frozen
		if self.hasBegun and bullettime.get() > 0 then
			if self.startTimer then
				self.scenarioTimeOut = self.scenarioTimeOut+dt*bullettime.get()
				--If truck(s) are not pending entering the scoreboard and above condition is false, apply wind to the player's car
			elseif self.pendTruck == 0 then
				local playerObject = scenetree.findObject('scenario_player0')
				playerObject:queueLuaCommand('obj:setWind(0, 150, 0)')
			end
		end
	end
end

function Curling:onBeamNGTrigger(data)
	--If a truck or player car has entered the scoreboard
	if data.triggerName == 'ice' and (data.subjectName == 'scenario_player0' or data.subjectName:find('truck')) and data.event == "enter" then
		--If the vehicles is a truck, this truck is no longer pending
		if data.subjectName:find('truck') then self.pendTruck = self.pendTruck-1 end
		--If there are no more trucks, let the timer run
		if self.pendTruck == 0 then
			self.startTimer = true
		end
		helper.queueLuaCommandByName('scenario_player0', 'bullettime.set(4)')
		helper.queueLuaCommandByName(data.subjectName, 'obj:setWind(0, 0, 0)')
	end
end

function Curling:triggerTruck()
	--[[
	Adds another truck to those pending to stop the scenario ending prematurely
	]]--
	self.pendTruck = self.pendTruck + 1
end

function Curling:forceScore(vehicle, score)
	--[[
	Forces an end score on a particular car (currently only used for Theme 4:2)
	]]--
	table.insert(self.forcedScores, {name = vehicle, score = score})
end

function Curling:onRaceStart()
	self.hasBegun = true
end

---------------------------------------------------------------------------------------------------------------------------------------------
---Class Cuegames - Inherits/Extends off of ClassEvent---
---------------------------------------------------------------------------------------------------------------------------------------------

Cuegames = {};
Cuegames.__index = Cuegames

function Cuegames:getBallWorth(ballName, theme)
	--[[
	Returns the number of points per ball
	]]--
	--Remove the ball number from the name
	local ballColour = ballName:gsub("[0-9]", "")
	if     ballColour == "red"    then return 100
	elseif ballColour == "yellow" then
		--Special case for theme 2
		if theme == 2 then return 0
		else return 200
		end
	elseif ballColour == "green"  then return 300
	elseif ballColour == "brown"  then return 400
	elseif ballColour == "blue"   then return 500
	elseif ballColour == "pink"   then return 600
	elseif ballColour == "black"  then
		--Special cases for themes 2 and 3
		if     theme == 2 then return 200
		elseif theme == 3 then return 700
		end
	end
end

function Cuegames:getOrder(theme)
	--[[
	Returns the order the balls need to be potted in, if applicable
	]]--
	if     theme == 2 then return {balls = {"red", "red", "red", "red", "red", "black"}, length = 6, current = 1}
	elseif theme == 3 then return {balls = {"red", "yellow", "green", "brown", "blue", "pink", "black"}, length = 7, current = 1}
	else return nil
	end
end

function Cuegames.new(eventType, filePath, scenarioName)
	local theme = getTheme(filePath)

	local maxScoreBall = 0
	local vehicles = {"scenario_player0"}
	local colours = {"red", "yellow", "green", "brown", "blue", "pink", "black"}
	for _, v in ipairs(colours) do
		local i = 0
		local name = v..i
		while scenetree.findObject(name) and i < 20 do
			--Add this ball into the vehicles so that they can trigger powerups
			table.insert(vehicles, name)
			maxScoreBall = maxScoreBall + Cuegames:getBallWorth(name, theme)
			i = i+1
			name = v..i
		end
	end

	local stats = event.getScenarioStats(theme, event.getEventNumber(eventType))

	--These are mandatory attributes
	local attributes = {
		scenarioName = scenarioName,
		eventName = eventType,
		theme = theme,
		scenarioTimeOut = 0,
		start = false,
		scoreBall = 0,
		lostBall = 0,
		ballsPotted = {},
		--Only themes 2 and 3 score time
		isScoringTime = theme == 2 or theme == 3,
		scenarioTimeLimit = stats.time.limit,
		scenarioTimeWorth = stats.time.worth,
		maxScoreBall = maxScoreBall,
		ballOrder = Cuegames:getOrder(theme),
		vehicles = vehicles
	}

	uiHandler.updateTimerUI(attributes.scenarioTimeLimit, attributes.scenarioTimeLimit, attributes.eventName)

	local instance = ClassEvent.new(attributes)

	setmetatable(instance, Cuegames)
	return instance
end

--Inheritance
setmetatable(Cuegames, {__index = ClassEvent})

function Cuegames:onRaceTick(raceTickTime)
	if not self.finished then
		if self.start then self.scenarioTimeOut = self.scenarioTimeOut+raceTickTime end

		--If the timer has reached the event time limit or all balls have been potted (that have not been destroyed), finish event
		if self.scenarioTimeOut >= self.scenarioTimeLimit or self.scoreBall == self.maxScoreBall-self.lostBall then
			local scoreTime = nil
			--If the event is scoring time, assign the variables
			if self.isScoringTime then
				scoreTime = math.floor(self.scenarioTimeLimit-self.scenarioTimeOut)*self.scenarioTimeWorth
			end
			event.finishScenario(self.eventName, self.theme, self.scenarioName, self.powerups, {scoreBall = self.scoreBall, scoreTime = scoreTime, lostBall = self.lostBall})
			self.finished = true
		end
	end
end

function Cuegames:destoryBall(ball)
	--[[
	Removes the ball from view and updates the variable for minusing later
	]]--
	TorqueScript.eval(''..ball..'.position = "10000 0 -10";')
	helper.flashUiMessage('A Ball Has Been Destroyed!', 3)
	self.lostBall = self.lostBall + 100
end

function Cuegames:ballsLeft()
	--[[
	Returns how many balls are left to pot
	]]--
	local balls = 0
	for _ in pairs(self.ballsPotted) do
		balls = balls+1
	end

	if self.ballOrder then return self.ballOrder.length-balls
	else return 10-balls
	end
end

function Cuegames:onRaceStart()
	--[[
	Removes the triangle from the start
	]]--
	TorqueScript.eval([[
		triangle_1.position = "10000 0 0";
		triangle_2.position = "10000 0 0";
		triangle_3.position = "10000 0 0";
	]])
	be:reloadStaticCollision();

	self.start = true
end

function Cuegames:onScenarioRestarted()
	--[[
	Resets the triangle
	]]--
	TorqueScript.eval([[
		triangle_1.position = "-563.864563 -1213.0249 29.0229187";
		triangle_2.position = "-555.798157 -1213.27441 29.0229187";
		triangle_3.position = "-559.747009 -1206.63586 29.0327263";
	]])
	be:reloadStaticCollision();

	--Calls the superclass for resetting powerups
	ClassEvent.onScenarioRestarted(self)
end

function Cuegames:onBeamNGTrigger(data)
	--Don't repeat when exiting trigger
	if data.event == 'enter' then
		if data.triggerName:find("pocket") then
			if data.subjectName ~= "scenario_player0" then
				local ballColour = data.subjectName:gsub("[0-9]", "")

				--If an incorrect ball has been potted, fail the scenario
				if self.ballOrder and self.ballOrder.balls[self.ballOrder.current] ~= ballColour then
					-- self:fail('Foul: Incorrect ball potted!| Expected: '..self.ballOrder.balls[self.ballOrder.current].. ' | Potted: '..data.subjectName:gsub("[0-9]", ""))
					self:fail("Foul: Incorrect ball potted!")
				else
					--Get worth and add to the current score
					local ballWorth = self:getBallWorth(data.subjectName, self.theme)
					self.scoreBall = self.scoreBall+ballWorth
					--Add the ball to those potted
					table.insert(self.ballsPotted, data.subjectName)

					if self.ballOrder and self.ballOrder.balls[self.ballOrder.current] == ballColour then
						--Update the next ball to be potted from table, provding there is one (if there is not, the scenario will end and it won't matter)
						if self.ballOrder.current+1 <= self.ballOrder.length then
							self.ballOrder.current = self.ballOrder.current+1
						end
					end
				end

			--Indicates player entered a pocket
			else
				self:fail("Foul: You drove into a pocket!")
			end
		elseif data.triggerName == 'outside' then
			local potted = false

			--Balls will activate this trigger when they fall to the bottom of a pocket
			--This determines whether they were potted and so should not trigger the failure of the scenario
			for key, value in ipairs(self.ballsPotted) do
				if value == data.subjectName then
					potted = true
				end
			end

			if not potted then
				self:fail('Foul: A ball flew off the table!')
			end
		end
	end
end

function Cuegames:onPreRender(dt)
	--Since this function has been overwritten, this is needed to access classEvent's
	--onPreRender function : ie. emulates super()
	ClassEvent.onPreRender(self, dt)

	if not self.finished then
		uiHandler.updateMedalUI(self.eventName, self.theme, self.powerups, true, nil, {scoreBall = self.scoreBall, lostBall = self.lostBall})
		uiHandler.updateTimerUI(self.scenarioTimeOut, self.scenarioTimeLimit, self.eventName)
	end
end

---------------------------------------------------------------------------------------------------------------------------------------------
---Class Targets - Inherits/Extends off of ClassEvent---
---------------------------------------------------------------------------------------------------------------------------------------------

Targets = {};
Targets.__index = Targets
function Targets.new(eventType, filePath, scenarioName, timeWorth, maxTimePoints, maxTargetPoints)
	local theme = getTheme(filePath)

	local stats = event.getScenarioStats(theme, event.getEventNumber(eventType))
	--These are mandatory attributes
	local attributes = {
		scenarioName = scenarioName,
		eventName = eventType,
		theme = theme,
		scoreTargets = 0,
		scenarioTime = 0,
		targets = {},
		timeLimit = stats.time.limit,
		timeWorth = stats.time.worth,
		maxTimePoints = stats.max.time,
		maxTargetPoints = stats.max.targets
	}

	uiHandler.updateTimerUI(0, stats.time.limit, eventType)

	local instance = ClassEvent.new(attributes)

	setmetatable(instance, Targets)
	return instance
end

--Inheritance
setmetatable(Targets, {__index = ClassEvent})

function Targets:onBeamNGTrigger(data)
	--Don't run agan when exiting trigger
	if data.event == "enter" then
		ClassEvent.onBeamNGTrigger(self, data)

		if data.triggerName == 'finish' then
			event.finishScenario(self.eventName, self.theme, self.scenarioName, self.powerups,  {targetsPoints = self.scoreTargets, scenarioTime = self.scenarioTime, timeWorth = self.timeWorth, timeLimit = self.timeLimit, maxTimePoints = self.maxTimePoints, maxTargetPoints = self.maxTargetPoints})
			self.finished = true
		else
			if string.find(data.triggerName, "target") then
				--Determines whether the player has already got points from this target
				local enteredBefore = false
				for key, value in ipairs(self.targets) do
					if value == data.triggerName then
						enteredBefore = true
					end
				end

				if not enteredBefore then
					local pointsValue = string.match(data.triggerName, '%d+')
					--If there are some points in the name of it (which there should ALWAYS be)
					if pointsValue ~= "" then
						local targetPts = tonumber(pointsValue)
						table.insert(self.targets, data.triggerName)
						self.scoreTargets = self.scoreTargets + targetPts
						uiHandler.updateMedalUI(self.eventName, self.theme, self.powerups, false, self.scoreTargets)
					else
						log("E", "CarGame", "A target has been incorrectly named in world! {TriggerName: "..data.triggerName.." | pointsValue: "..pointsValue)
					end
				end
			end
		end
	end
end

function Targets:onRaceTick(raceTickTime)
	if not self.finished then
		self.scenarioTime = self.scenarioTime + raceTickTime
		uiHandler.updateTimerUI(self.scenarioTime, self.timeLimit, self.eventName)
	end
end

---------------------------------------------------------------------------------------------------------------------------------------------
---Class Airborne - Inherits/Extends off of ClassEvent---
---------------------------------------------------------------------------------------------------------------------------------------------

Airborne = {};
Airborne.__index = Airborne
function Airborne.new(eventType, filePath, scenarioName)
	local scoreboard = scenetree.findObject("scoreboard"):getPosition()
	local floorData = {floor = scoreboard.z + 4, yCoord = -scoreboard.y}

	local theme = getTheme(filePath)

	local stats = event.getScenarioStats(theme, event.getEventNumber(eventType))

	--These are mandatory attributes
	local attributes = {
		scenarioName = scenarioName,
		eventName = eventType,
		finished = false,
		theme = theme,
		active = false,
		nitro = 0,
		maxSpeedAchieved = nil,
		floorData = floorData,
		--Only theme 5 has speed scoring
		isScoringSpeed = theme == 5
	}

	local instance = ClassEvent.new(attributes)

	uiHandler.updateMedalUI(eventType, instance.theme, instance.powerups, false, nil, 0)

	setmetatable(instance, Airborne)
	return instance
end

--Inheritance
setmetatable(Airborne, {__index = ClassEvent})

function Airborne:onBeamNGTrigger(data)
	--Again, do not run twice
	if data.event == "enter" then
		ClassEvent.onBeamNGTrigger(self, data)
		--Activates nitro at the end of the ramp. IF no nitro is in the event, this has no effect
		if data.triggerName == 'ramp' then
			self.active = true
		end
	end
end

function Airborne:onRaceTick(raceTickTime)
	if not self.finished then
		local playerObject = scenetree.findObject("scenario_player0")
		local distance = playerObject:getPosition().y + self.floorData.yCoord

		--If the player's car is below the height of the scoreboard, finish event
		if playerObject:getPosition().z <= self.floorData.floor then
			event.finishScenario(self.eventName, self.theme, self.scenarioName, self.powerups, {distance = math.max(0, distance), speed = self.maxSpeedAchieved})
			self.finished = true
		end

		--If the player has not reached the ramp and speed scoring is active, then update the player's max score if required
		if not self.active and self.isScoringSpeed then
			local playerData = map.objects[map.objectNames["scenario_player0"]]

			--*10 relates to *10 points per ms^-1 of velocity
			--Ie, 100ms^-1 relates to 1000 points
			local speed = playerData.vel:length()*10
			if not self.maxSpeedAchieved or speed > self.maxSpeedAchieved then
				self.maxSpeedAchieved = speed
			end
		end

		--If powerups are in the event
		if self.powerups ~= nil then
			--If the player has reached the ramp and collected nitro powerups
			if self.active and self.nitro > 0 then
				playerObject:queueLuaCommand('obj:setWind(0, 300, 0)')
				self:addNitro(-raceTickTime)
			--If the player has run out of nitro, set the wind to normal
			elseif self.nitro <= 0 then
				playerObject:queueLuaCommand('obj:setWind(0, 0, 0)')
				self.active = false
			end
		end
	end
end

---------------------------------------------------------------------------------------------------------------------------------------------

local function createNewEvent(filePath, ...)
	--[[
	Instantiates the correct class for use in the scenario
	]]--
	local eventType = file.getEventType(file.determineEventFromFile(filePath))

	--Get scenario name
	local scenarioData = jsonReadFile(filePath)[1]
	local scenarioName = scenarioData.name

	if not eventType then
		log("E", "Car Games", "Event was setup without a correct event type! Loading has been terminated")
    return
	end

	if     eventType == "darts"    then scenarioInstance = Darts   .new(eventType, filePath, scenarioName)
	elseif eventType == "curling"  then scenarioInstance = Curling .new(eventType, filePath, scenarioName)
	elseif eventType == "cuegames" then scenarioInstance = Cuegames.new(eventType, filePath, scenarioName, ...)
	elseif eventType == "targets"  then scenarioInstance = Targets .new(eventType, filePath, scenarioName, ...)
	elseif eventType == "airborne" then scenarioInstance = Airborne.new(eventType, filePath, scenarioName, ...)
	end
end

local function updateEvent(funcName, var, ...)
	--[[
	Updates the event with the various functions to make the scenario work
	]]--
	if not scenarioInstance then return end

	if funcName == "onPreRender" then
		--This is used to override onPreRender which can cause problems when changing scenarios
		--This is a fix of the root cause of the 'obj not found' errors when finding the player car
		if not scenarioInstance:isFinished() then
			scenarioInstance:onPreRender(var)
		end
	else
		scenarioInstance:callMethod(funcName, var, ...)
	end
end

local function callMethod(funcName, var, ...)
	--[[
	Calls and returns a function from the scenarioInstance
	]]--
	return scenarioInstance:callMethod(funcName, var, ...)
end

M.callMethod = callMethod
M.createNewEvent = createNewEvent
M.updateEvent = updateEvent
return M
