local M = {}

local statistics          = require('scenario/cargames/statisticsHandler')
local championshipHandler = require('scenario/cargames/championshipHandler')
local fileHandler         = require('scenario/cargames/fileHandler')
local helper              = require('scenario/scenariohelper')

local function getMedalScore(theme, event)
  --[[
  Returns the scores needed for each medal
  These are hardcoded values and should NOT be changed
  ]]--
  if theme == nil or event == nil then return end

  if theme == 1 then
    if     event == 1 then return {100, 150, 160, 170, 180}
    elseif event == 2 then return {145, 155, 165, 175, 180}
    elseif event == 3 then return {400, 500, 600, 700, 800}
    elseif event == 4 then return {100, 150, 190, 210, 250}
    elseif event == 5 then return {250, 300, 450, 500, 600}
    end
  elseif theme == 2 then
    if     event == 1 then return {100, 200, 350, 800, 1100}
    elseif event == 2 then return {100, 800, 3750, 4800, 6000}
    elseif event == 3 then return {300, 500, 700, 1000, 1350}
    elseif event == 4 then return {200, 300, 400, 410, 420}
    elseif event == 5 then return {500, 600, 750, 1100, 1350}
    end
  elseif theme == 3 then
    --Theme 3 scores have been revised for bronze and silver so unlocking the
    --next theme is not as difficult
    if     event == 1 then return {100, 300, 1000, 2500, 11000}
    elseif event == 2 then return {10, 200, 750, 875, 1000}
    elseif event == 3 then return {1250, 2100, 3600, 3800, 4000}
    elseif event == 4 then return {400, 600, 950, 1000, 1030}
    elseif event == 5 then return {400, 650, 1150, 1300, 1400}
    end
  elseif theme == 4 then
    if     event == 1 then return {300, 1000, 2000, 4000, 8000}
    elseif event == 2 then return {4000, 12000, 24000, 40960, 81920}
    elseif event == 3 then return {3000, 4000, 5000, 6000, 9750}
    elseif event == 4 then return {600, 800, 950, 1000, 1200}
    elseif event == 5 then return {200, 400, 500, 575, 625}
    end
  elseif theme == 5 then
    if     event == 1 then return {75, 100, 150, 200, 300}
    elseif event == 2 then return {75, 100, 150, 160, 170}
    elseif event == 3 then return {400, 500, 700, 800, 1000}
    elseif event == 4 then return {1000, 1500, 2000, 2250, 2500}
    elseif event == 5 then return {500, 850, 1100, 1400, 1595}
    end
  end
end

local function getScenarioStats(theme, event)
  --[[
  Returns the scenario stats such as timing and distances
  These are hardcoded values and should NOT be changed
  ]]--
  if theme == 1 then
    if     event == 3 then return {time = {limit = 90, worth = 0}}
    elseif event == 4 then return {time = {limit = 60, worth = 1}, max = {time = 25, targets = 320}}
    elseif event == 5 then return {max = {distance = 640}}
    end
  elseif theme == 2 then
    if     event == 3 then return {time = {limit = 80, worth = 50}, max = {time = 2000}}
    elseif event == 4 then return {time = {limit = 60, worth = 1}, max = {time = 35, targets = 400}}
    elseif event == 5 then return {max = {distance = 1400}}
    end
  elseif theme == 3 then
    if     event == 3 then return {time = {limit = 90, worth = 50}, max = {time = 1500}}
    elseif event == 4 then return {time = {limit = 150, worth = 10}, max = {time = 350, targets = 760}}
    elseif event == 5 then return {max = {distance = 1500}}
    end
  elseif theme == 4 then
    if     event == 4 then return {time = {limit = 135, worth = 10}, max = {time = 500, targets = 880}}
    elseif event == 5 then return {max = {distance = 780}}
    end
  elseif theme == 5 then
    if     event == 3 then return {time = {limit = 90, worth = 0}}
    elseif event == 4 then return {time = {limit = 600, worth = 3}, max = {time = 1250, targets = 1560}}
    elseif event == 5 then return {max = {distance = 800, speed = 1100}}
    end
  end
end

local function getEventNumber(event)
  --[[
  Returns the event number for a given event
  ]]--
  if event == nil then return end

  if     event == "darts"    then return 1
  elseif event == "curling"  then return 2
  elseif event == "cuegames" then return 3
  elseif event == "targets"  then return 4
  elseif event == "airborne" then return 5
  end
end

local function calculateMedal(medalScores, score)
  --[[
  Determines what medal to award the player
  ]]--
  if medalScores == nil or score == nil then return end

  local medals = {"bronze", "silver", "gold", "platinum", "diamond"}
  local medal = ""

  for i = 1, 5 do
    if score < medalScores[i] then
      if i == 1 then
        --No medal has been achieved
        medal = "none"
      else
        --Award previous medal
        medal = medals[i-1]
      end

      break
    elseif i == 5 then
      --Award diamond
      medal = medals[5]
    end
  end

  return medal
end

local function getNextEvent(currentEventNum, themeNum)
  --[[
  Returns the next event
  ]]--
  if currentEventNum < 5 then
    local scenarioFiles = fileHandler.findAllFiles()
    return scenarioFiles[currentEventNum + 1][themeNum]:sub(2)
  end

  return nil
end

local function showEndScreen(resultsData, medalData, scenarioData)
	--[[
	Displays the end screen UI
	]]--
  bullettime.set(1/8)
	guihooks.trigger("carGamesEndScreen", {resultsData = resultsData, medalData = medalData, scenarioData = scenarioData})
end

local function endScenario(scenarioData)
  --[[
  Ends the scenario and setups up the end screen UI with the necessary information
  ]]--
  if not scenarioData then
    log("E", "Car Games: endScenario()", "scenarioData is nil")
    return
  end

  --Convert to a number
  scenarioData.event = fileHandler.determineEventFromFile(scenarioData.event)

  local medalScores = getMedalScore(scenarioData.theme, scenarioData.event)
  local medalAchieved = calculateMedal(medalScores, scenarioData.scores and scenarioData.scores.total or 0)
  local medalAchievedWorth = fileHandler.getMedalRanking(medalAchieved)

  local isInChapionshipMode = championshipHandler.isInChapionshipMode()

  --The three tables required for the UI
  local resultsData, medalInfo, scenarioInfo = {}, {}, {}

  scenarioInfo = {
    theme = scenarioData.theme,
    isInChapionshipMode = isInChapionshipMode,
    championshipData = {},
    name = scenarioData.name,
    nextScenario = getNextEvent(scenarioData.event, scenarioData.theme)
  }

  if not scenarioData.failed then
    scenarioInfo.status = {success = true, reason = ""}

    local totalRow = {}
    local multipliersRow = nil
    for k, v in pairs(scenarioData.scores) do
      local reason = k
      if reason == "multipliers" then
        --Don't add in multiplier if there are no multipliers
        if #v == 0 then
          goto continue
        else
          table.sort(v)
          multipliersRow = {amount = v, reason = "Multiplier"}
          goto continue
        end
      --multiplier is the number, not the table that has been accounted for above
      elseif reason == "multiplier" then
        goto continue
      else
        --Space between capital letters
        reason = reason:gsub("%u", " %1")
        --Uppercase first letter
        reason = reason:sub(1,1):upper()..reason:sub(2)
        --Remove 'Score ' from the name
        reason = reason:gsub("Score ", "")
      end

      ::addToTable::
      local data = {amount = v, reason = reason}

      if reason == "Total" then
        total = data
        goto continue
      end

      table.insert(resultsData, data)

      ::continue::
    end

    --Insert the multiplier row above the total row but after the rest
    if multipliersRow then
      table.insert(resultsData, multipliersRow)
    end
    --Insert the total row into the end of the data
    table.insert(resultsData, total)

    local medals = {"bronze", "silver", "gold", "platinum", "diamond"}

    local currentMedal = fileHandler.getAttributeFromSave({theme = scenarioData.theme, event = scenarioData.event, value = "medal"})
    local medalWorth = fileHandler.getMedalRanking(currentMedal)

    --Add each medal to the list with their score and whether it has been surpassed
    for i, v in ipairs(medalScores) do
      medalInfo[medals[i]] = {score = v, gotten = medalWorth >= i}
    end
    --Add in best score data as well
    local bestScore = fileHandler.getAttributeFromSave({theme = scenarioData.theme, event = scenarioData.event, value = "score"})
    medalInfo.best = {score = bestScore, gotten = scenarioData.scores.total >= bestScore, newMedal = medalAchievedWorth > medalWorth}

    --Must happen after above so that the new can be compared to the previous best
    fileHandler.setAttributeToSave({theme = scenarioData.theme, event = scenarioData.event, name = "medal", value = medalAchieved})
    fileHandler.setAttributeToSave({theme = scenarioData.theme, event = scenarioData.event, name = "score", value = scenarioData.scores.total})
  else
    scenarioInfo.status = {success = false, reason = scenarioData.msg}
  end

  if not isInChapionshipMode then
    --Only set last event info if a stand-alone event
    fileHandler.setLastScenarioInfo({theme = scenarioData.theme, event = scenarioData.event})
  else
    scenarioData.medal = medalAchievedWorth

    local finished, champData = championshipHandler.endEvent(scenarioData)

    scenarioInfo.championshipData = {
      {title = "Event Points", value = champData.eventScores},
      {title = "Total Points", value = champData.overallScore}
    }

    if finished then
      scenarioInfo.championshipData[3] = {status = "Result", msg = champData.overallScore > 75 and "Success" or "Failure"}
    end
  end

  showEndScreen(resultsData, medalInfo, scenarioInfo)
end

local function getScores(finished, scData, powerups, ...)
  --[[
  Returns the scores for an event, given the scenario's data (powerups, eventType .etc)
  ]]--
  scData.scores = {}
  scData.scores.multiplier = 1
  scData.scores.multipliers = {}

  local eventNumber = fileHandler.determineEventFromFile(scData.event)

  if powerups ~= nil then
    --Multipliers are displayed to the user at the end screen and so need to be counted
    for k, v in pairs(powerups.instances) do
      --Multipliers have values in them but also need to ignore time based powerups
      --since they activate during the event
      if v.name:find("%d+") and not v.name:find("time") then
        local value = tonumber(v.name:match("%d+"))
        if v:hasBeenTriggered() then
          scData.scores.multiplier = scData.scores.multiplier*value
          table.insert(scData.scores.multipliers, value)
        end
      end
    end

    scData.powerups = powerups
  end

  if scData.event == "darts" then
    local scores = ...

    local scoreLand = math.max(0, math.floor(scores.scoreLand))
    local scoreRest = math.max(0, math.floor(scores.scoreRest))
    local scoreFinal = math.max(0, scoreLand+scoreRest)

    scData.scores.scoreFinalPosition = scoreRest
    scData.scores.scoreLand = scoreLand
    scData.scores.total = math.floor(scoreFinal*scData.scores.multiplier)
  elseif scData.event == "curling" then
    local objects = select(1, ...)
    local forcedScores = select(2, ...)
    local scores = {}

    for _, v in ipairs(objects) do
      local added = false
      if forcedScores ~= nil then
        --forces scores on vehicles
        for _, u in ipairs(forcedScores) do
          if v == u.name then
            table.insert(scores, {vehicleName = v, distance = 100-u.score})
            added = true
            break
          end
        end
      end

      --If the vehicle has already had its score force, don't bother working it out
      if not added then
        if not v:find("truck") then
          local obj = scenetree.findObject(v)
          --Note:  these values (563.3, 0.01, 733.6) are hardcoded due to the object in scenario not having its center as the scoreboard center
          local distance = math.sqrt(math.pow(obj:getPosition().x+563.3, 2) + math.pow(obj:getPosition().y+733.6, 2), math.pow(obj:getPosition().z+0.01, 2))
          table.insert(scores, {vehicleName = v, distance = distance})
        end
      end
    end

    local scoreFinal = 0
    for _, carData in pairs(scores) do
      local score = math.max(0, math.floor(100 - carData.distance*2))
      local vehName = ""

      if not carData.vehicleName:find("scenario_player") then
        --Vehicles that are not called "scenario_player<%d+>" are opponents, so subtract score
        score = -score
        --+1 since opponent names start at 0
        vehName = "Opponent "..tonumber(carData.vehicleName:match("%d+")+1)
      else
        if carData.vehicleName == "scenario_player0" then
          vehName = "Player"
        else
          vehName = "Teammate "..carData.vehicleName:match("%d+")
        end
      end

      scoreFinal = scoreFinal + score
      scData.scores[vehName] = score
    end

    scoreFinal = math.max(0, scoreFinal)
    scData.scores.total = math.floor(scoreFinal*scData.scores.multiplier)
  elseif scData.event == "cuegames"  then
    local vars = ...
    if vars == nil then return end

    scData.scores.scoreBallsPotted = vars.scoreBall

    --Only apply time score if it is being scored
    if not vars.scoreTime then
      scData.scores.total = scData.scores.scoreBallsPotted
    else
      scData.scores.scoreTime = math.floor(vars.scoreTime)
      scData.scores.total = scData.scores.scoreBallsPotted + scData.scores.scoreTime
    end

    if vars.lostBall and vars.lostBall > 0 then
      scData.scores.ballsLost = -vars.lostBall
      scData.scores.total = scData.scores.total + scData.scores.ballsLost
    end

    if vars.scoreStorms then
      scData.scores.scoreStorms = vars.scoreStorms
      scData.scores.total = scData.scores.total + scData.scores.scoreStorms
    end
  elseif scData.event == "targets" then
    local vars = ...
    if vars == nil then return end

    scData.scores.scoreTargets = vars.targetsPoints
    scData.scores.scoreTime = math.floor((vars.timeLimit-vars.scenarioTime)*vars.timeWorth)

    scData.scores.total = math.max(0, math.floor(scData.scores.scoreTargets+scData.scores.scoreTime))
  elseif scData.event == "airborne" then
    local vars = ...
    if vars == nil then return end

    --Don't want negative scores
    scData.scores.distance = math.max(0, math.floor(vars.distance))

    if vars.speed == nil then
      scData.scores.total = scData.scores.distance
    else
      scData.scores.speed = math.floor(vars.speed)
      scData.scores.total = scData.scores.distance + scData.scores.speed
    end
  end

  return scData
end

local function finishScenario(event, theme, name, powerups, ...)
  --[[
  Finishes the scenario - used as a way to call multiple other functions compactly and in
  an organised fashion
  ]]--
  if event == nil or theme == nil then return end

  local scData = {}
  scData.event = event
  scData.theme = theme
  scData.name = name

  scData = getScores(true, scData, powerups, ...)

  endScenario(scData)
end

M.getScores = getScores
M.getScenarioStats = getScenarioStats
M.getEventNumber = getEventNumber
M.getNextEvent = getNextEvent
M.finishScenario = finishScenario
M.endScenario = endScenario
M.getMedalScore = getMedalScore
return M
