local M = {}

local function getTotalNumberOfScenarios()
	--[[
	Returns the total number of scenarios - this value is hardcoded
	]]--
	return 25
end

local function getTotalNumberOfThemes()
	--[[
	Returns the number of themes
	]]--
	return getTotalNumberOfScenarios()/5
end

local function createNewSave()
	--[[
	Creates a blank new save file
	]]--
	local content = {}
	for i = 1, getTotalNumberOfThemes() do
		content["theme"..i] = {}
		for j = 1, 5 do
			content["theme"..i]["event"..j] = {
				medal = "none",
				score = 0
			}
		end
	end
	content["lastScenario"] = {
		theme = 0,
		event = 0,
		championship = 0
	}
	content["championship"] = {
		engaged = false,
		data = {
			event = 0,
			theme = 0,
			overallScore = 0,
			eventScores = {}
		},
		themes = {
			0,
			0,
			0,
			0,
			0
		}
	}

	serializeJsonToFile("cargames_savefile", content, true)
end

local function getEventType(num)
	--[[
	Gets the name of the event from the number
	]]--
	if     num == 1 then return "darts"
	elseif num == 2 then return "curling"
	elseif num == 3 then return "cuegames"
	elseif num == 4 then return "targets"
	elseif num == 5 then return "airborne"
	end
end

local function determineEventFromFile(e)
	--[[
	Takes a fileName, e, and returns the event number
	]]
	if     e:find("darts")    then return 1
	elseif e:find("curling")  then return 2
	elseif e:find("cuegames") then return 3
	elseif e:find("targets")  then return 4
	elseif e:find("airborne") then return 5
	end
end

local function getThemeName(num)
	--[[
	Gets the theme name from the number
	]]--
	if     num == 1 then return "Forest"
	elseif num == 2 then return "Technology"
	elseif num == 3 then return "Space"
	elseif num == 4 then return "Planets"
	elseif num == 5 then return "Industrial"
	end
end

local function findAllFiles()
	--[[
	Locates and finds all JSON files from the mod
	]]--
	--In the format allFiles[<event>][<theme>]
	local allFiles = {}
	for i = 1, 5 do
		allFiles[i] = {}
		for j = 1, getTotalNumberOfThemes() do
			allFiles[i][j] = nil
		end
	end

	local files1 = FS:findFilesByPattern('/levels/smallgrid/scenarios/cargames/themes/1_forest/'    , '*.json', 1, true, false)
	local files2 = FS:findFilesByPattern('/levels/smallgrid/scenarios/cargames/themes/2_tech/'      , '*.json', 1, true, false)
	local files3 = FS:findFilesByPattern('/levels/smallgridspace/scenarios/cargames/themes/3_space/', '*.json', 1, true, false)
	local files4 = FS:findFilesByPattern('/levels/smallgrid/scenarios/cargames/themes/4_planets/'   , '*.json', 1, true, false)
	local files5 = FS:findFilesByPattern('/levels/smallgrid/scenarios/cargames/themes/5_industrial/', '*.json', 1, true, false)
	local files = {files1, files2, files3, files4, files5}

	for theme, file in ipairs(files) do
		for _, eventName in pairs(file) do
			local event = determineEventFromFile(eventName)
			--Check that the event exists (ie. is not something like cargameshub, which is not an event)
			if event then
				allFiles[event][theme] = eventName
			end
		end
	end

	return allFiles
end

local function getMedalName(medal)
	--[[
	Returns the name of the medal from the medal number
	]]--
	if not medal then return end

	if     medal == 5 then return "diamond"
	elseif medal == 4 then return "platinum"
	elseif medal == 3 then return "gold"
	elseif medal == 2 then return "silver"
	elseif medal == 1 then return "bronze"
	elseif medal == 0 then return "none"
	else return nil
	end
end

local function checkFileExists()
	--[[
	Checks to see if it can open the savefile. If it can then also return the file
	so any function calling this doesn't have to open the file again - efficiency
	]]--
	local file = readFile("cargames_savefile")
	if file then
		return true, file
	else
		return false
	end
end

local function readFromFile()
	--[[
	Checks the file exists and it up to date, and then reads the file
	]]--
	local fileExists, file = checkFileExists()
	if not fileExists then
		createNewSave()
		--Update the file variable with newly created info
		_, file = checkFileExists()
	end

	return jsonDecode(file, "cargames_savefile")
end

local function getMedalsAchieved(theme)
	--[[
	The parameter theme is optional, and specified will only return the medals
	achieved for that specific theme
	Otherwise, returns the total number of each medal achieved
	]]--
	local file = readFromFile()
	local medals = {bronze = 0, silver = 0, gold = 0, platinum = 0, diamond = 0, none = 0}
	if theme == nil then
		for t = 1, getTotalNumberOfThemes() do
			for e = 1, 5 do
				local medalAchieved = file["theme"..t]["event"..e]["medal"]
				medals[medalAchieved] = medals[medalAchieved]+1
			end
		end
	else
		for e = 1, 5 do
			local medalAchieved = file["theme"..theme]["event"..e]["medal"]
			medals[medalAchieved] = medals[medalAchieved]+1
		end
	end

	return medals
end

local function getMedalRanking(medal)
	--[[
	Gets the medal rank/number from the medal name
	]]--
	--Makes medal is case correct for if-else ladder below
	medal = medal:lower()

	if     medal == "diamond"  then return 5
	elseif medal == "platinum" then return 4
	elseif medal == "gold"     then return 3
	elseif medal == "silver"   then return 2
	elseif medal == "bronze"   then return 1
	elseif medal == "none"     then return 0
	else return nil
	end
end

local function getAttributeFromSave(attributeData)
	--[[
	Returns a specific attribute for a specific scenario
	]]--
	local file = readFromFile()
	if file["theme"..attributeData.theme]["event"..attributeData.event][attributeData.value] == nil then return end
	return file["theme"..attributeData.theme]["event"..attributeData.event][attributeData.value]
end

local function getMiscAttribute(attribute)
	--[[
	Returns a miscellaneous attribute from the savefile
	]]--
	local file = readFromFile()
	return file[attribute]
end

local function getTotalNumberOfMedalsAchieved()
	--[[
	Returns the number of total medals achieved
	ie. a diamond medal is 5 medals
	]]--
	local medalsAchieved = 0
	for t = 1, getTotalNumberOfThemes() do
		for e = 1, 5 do
			local medal = getAttributeFromSave({value = "medal", theme = t, event = e})
			if medal ~= "none" then
				medalsAchieved = medalsAchieved + getMedalRanking(medal)
			end
		end
	end

	return medalsAchieved
end

local function setChampionshipInfo(championshipData, newState)
	--[[
	Sets the championship data to the file and overwrites any older data
	]]--
	local file = readFromFile()
	if championshipData then
		if championshipData.data == nil then
			file.championship.data = championshipData
		else
			file.championship = championshipData
		end
	end

	if newState ~= nil then
		file.championship.engaged = newState
	end
	
	serializeJsonToFile("cargames_savefile", file, true)
end

local function setLastScenarioInfo(attributeData)
	--[[
	Sets the last scenario exitied into the savefile for use in the hub-world
	]]--
	local file = readFromFile()
	if attributeData.championship then
		file["lastScenario"]["theme"] = 0
		file["lastScenario"]["event"] = 0
		file["lastScenario"]["championship"] = attributeData.championship
	else
		file["lastScenario"]["theme"] = attributeData.theme
		file["lastScenario"]["event"] = attributeData.event
	end
	serializeJsonToFile("cargames_savefile", file, true)
end

local function setAttributeToSave(attributeData)
	--[[
	Sets an attribute to the savefile for a specific scenario
	]]--
	local file = readFromFile()
	--Check that the attribute exists in the file and parameter
	if attributeData == nil then return end

	if attributeData.name == "medal" then
		attributeData.value = attributeData.value:lower()
		--Get current medal from file
		local currentMedal = getAttributeFromSave({value = "medal", theme = attributeData.theme, event = attributeData.event})
		--If the file has a lower ranked medal, update
		if getMedalRanking(currentMedal) < getMedalRanking(attributeData.value) then
			file["theme"..attributeData.theme]["event"..attributeData.event][attributeData.name] = attributeData.value
			serializeJsonToFile("cargames_savefile", file, true)
		end
	elseif attributeData.name == "score" then
		local currentScore = getAttributeFromSave({value = "score", theme = attributeData.theme, event = attributeData.event})
		if currentScore < attributeData.value then
			file["theme"..attributeData.theme]["event"..attributeData.event][attributeData.name] = attributeData.value
			serializeJsonToFile("cargames_savefile", file, true)
		end
	elseif attributeData.name == "gem" then
		file["theme"..attributeData.theme]["event"..attributeData.event][attributeData.name] = attributeData.value
		serializeJsonToFile("cargames_savefile", file, true)
	end
end

M.setAttributeToSave = setAttributeToSave
M.determineEventFromFile = determineEventFromFile
M.createNewSave = createNewSave
M.getMedalName = getMedalName
M.getMedalsAchieved = getMedalsAchieved
M.getMedalRanking = getMedalRanking
M.getAttributeFromSave = getAttributeFromSave
M.getTotalNumberOfMedalsAchieved = getTotalNumberOfMedalsAchieved
M.getTotalNumberOfScenarios = getTotalNumberOfScenarios
M.getTotalNumberOfThemes = getTotalNumberOfThemes
M.getMiscAttribute = getMiscAttribute
M.setLastScenarioInfo = setLastScenarioInfo
M.findAllFiles = findAllFiles
M.setChampionshipInfo = setChampionshipInfo
M.getThemeName = getThemeName
M.readFromFile = readFromFile
M.getEventType = getEventType
return M
